/*
 * Decode ABC80 cassette files
 *
 * Convert to single channel 16-bit with appropriate endianness first:
 *
 * sox muzak.wav --endian little -c 1 -s muzak.raw
 */

#include "abcdisk.h"
#include <math.h>

static unsigned int xtime;

static void output_block(uint8_t * buf, uint8_t * csum)
{
    uint16_t csum1, csum2;
    int i;
    static FILE *f;
    static uint16_t block;
    static char filename[64];
    int err = 0;

    if (!buf) {
        err = 1;                /* Notification of a bad block */
    } else {
        csum2 = csum[0] + (csum[1] << 8);

        csum1 = 0x03;           /* Checksum includes ETX */
        for (i = 0; i < 256; i++)
            csum1 += buf[i];

        if (csum1 == csum2) {
            if (!memcmp(buf, "\xff\xff\xff", 3)) {
                if (f)
                    fclose(f);

                unmangle_filename(filename, buf + 3);
                f = fopen(filename, "wb");
                if (!f) {
                    perror(filename);
                    exit(1);
                }
                block = 0;
                printf("File: %s\n", filename);
            } else if (!f || buf[0] != 0x00 ||
                       (buf[1] + (buf[2] << 8)) != block) {
                printf("Bad block ID: %02x %02x %02x\n",
                       buf[0], buf[1], buf[2]);
                err = 1;
            } else {
                block++;
                fwrite(buf + 3, 1, 253, f);
            }
        } else {
            printf("Checksum error: got %04x data %04x\n", csum2, csum1);
            err = 1;
        }
    }

    if (err && f) {
        fclose(f);
        remove(filename);
        f = NULL;
    }
}

static void output_bit(unsigned int bit)
{
    enum state {
        st_leader,
        st_sync,
        st_stx,
        st_data,
        st_etx,
        st_csum
    };
    static enum state st = st_leader;
    static unsigned int ctr = 0;
    static unsigned int bits = 0;
    static uint8_t data = 0;
    static uint8_t buf[256];
    static uint8_t csum[2];

    data = (data >> 1) + (bit << 7);    /* Bigendian bit order... */
    bits++;

    switch (st) {
    case st_leader:
        ctr++;
        if (bit == 1)
            ctr = 0;
        if (ctr > 128)          /* After 128 zeroes start looking for sync */
            st = st_sync;
        break;

    case st_sync:
        if (data == 0x16) {
            st = st_stx;        /* Sync acquired */
            bits = 0;
        }
        break;

    case st_stx:
        if (bits < 8)
            break;
        bits = 0;
        ctr = 0;
        if (data == 0x02) {
            st = st_data;
        } else if (data == 0x16) {
            /* Got another SYNC */
        } else {
            printf("Got %02x when expecting SYNC or STX\n", data);
            output_block(NULL, NULL);
            st = st_leader;     /* ERROR */
        }
        break;

    case st_data:
        if (bits < 8)
            break;
        bits = 0;
        buf[ctr++] = data;
        if (ctr == 256)
            st = st_etx;
        break;

    case st_etx:
        if (bits < 8)
            break;
        bits = 0;
        ctr = 0;
        if (data == 0x03)
            st = st_csum;
        else {
            printf("Got %02x when expecting ETX\n", data);
            output_block(NULL, NULL);
            st = st_leader;     /* ERROR */
        }
        break;

    case st_csum:
        if (bits < 8)
            break;
        bits = 0;
        csum[ctr++] = data;
        if (ctr == 2) {
            output_block(buf, csum);
            ctr = 0;
            st = st_leader;
        }
        break;
    }
}

static void process_flank(unsigned int time)
{
    static unsigned int since_start = 0;
    static unsigned int captured = 0;

    since_start += time;

    if (since_start < xtime / 4) {
        printf("Quick flank: %u:%u\n", since_start, xtime);
    } else if (since_start < xtime * 3 / 4) {
        captured = 1;
    } else if (since_start < xtime * 5 / 4) {
        output_bit(captured);
        since_start = captured = 0;
    } else {
        printf("Slow flank: %u:%u\n", since_start, xtime);
        since_start = 0;
    }
}

int main(int argc, char *argv[])
{
    unsigned int frequency;
    FILE *f;
    int16_t samp, osamp;
    double v, k, cutoff, thresh;
    unsigned int state;
    unsigned int time, ptime;

    setlocale(LC_ALL, "");

    if (argc < 2 || (argv[1][0] == '-' && !argv[1][1])) {
	fprintf(stderr, "abcdisk %s\n", VERSION);
	fprintf(stderr, "Usage: %s input.wav [samplerate [threshold_hz [cutoff]]]\n", argv[0]);
	exit(1);
    }

    f = fopen(argv[1], "rb");
    if (!f) {
	perror(argv[1]);
	exit(1);
    }

    frequency = (argc > 2) ? strtoul(argv[2], NULL, 0) : 44100;
    thresh = (argc > 3) ? atof(argv[3]) : 5300.0;
    cutoff = (argc > 4) ? atof(argv[4]) : 0.0;

    /* XXX: For ABC800 tapes, need to switch to 2400 baud after header */
    k = exp(-cutoff / frequency);
    xtime = frequency / 700;    /* Samples per baud */

    state = time = ptime = 0;
    samp = osamp = 0;
    v = 0.0;

    while (fread(&samp, 1, 2, f) == 2) {
        /* Emulate high pass filter in ABC80 */
        v = (v * k) + (samp - osamp);
        osamp = samp;

        if (v >= thresh) {
            if (!state) {
                state = 1;
                process_flank(time - ptime);
                ptime = time;
            }
        } else if (v <= -thresh) {
            if (state) {
                state = 0;
                process_flank(time - ptime);
                ptime = time;
            }
        }
        time++;
    }

    fclose(f);

    return 0;
}
